from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict

import aioredis

from eventiq.broker import Broker

from .settings import RedisSettings

if TYPE_CHECKING:
    from eventiq import CloudEvent, Consumer, RawMessage, Service


class RedisBroker(Broker[Dict[str, str]]):
    """
    Broker implementation based on redis PUB/SUB and aioredis package
    :param url: connection string to redis
    :param connect_options: additional connection options passed to aioredis.from_url
    :param kwargs: base class arguments
    """

    WILDCARD_ONE = "*"
    WILDCARD_MANY = "*"
    Settings = RedisSettings

    def __init__(
        self,
        *,
        url: str,
        connect_options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:

        super().__init__(**kwargs)
        self.url = url
        self.connect_options = connect_options or {}
        self._redis = None

    def parse_incoming_message(self, message: RawMessage) -> Any:
        return message

    @property
    def is_connected(self) -> bool:
        return self.redis.connection.is_connected

    async def _start_consumer(self, service: Service, consumer: Consumer) -> None:
        handler = self.get_handler(service, consumer)
        psub = self.redis.pubsub()
        while self._running:
            message = await psub.get_message(ignore_subscribe_messages=True)
            if message:
                await handler(message)

    async def _disconnect(self) -> None:
        await self.redis.close()

    @property
    def redis(self) -> aioredis.Redis:
        assert self._redis is not None, "Not connected"
        return self._redis

    async def _connect(self) -> None:
        self._redis = aioredis.from_url(url=self.url, **self.connect_options)

    async def _publish(self, message: CloudEvent, **kwargs) -> None:
        data = self.encoder.encode(message)
        await self.redis.publish(message.topic, data)



from __future__ import annotations

import asyncio
import os
import re
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Callable, Coroutine, Generic

import anyio
from pydantic import ValidationError

from .context import _current_message, _current_service
from .exceptions import DecodeError, Fail, Skip
from .imports import import_from_string
from .logger import LoggerMixin
from .message import Message
from .middleware import Middleware
from .models import CloudEvent
from .settings import BrokerSettings
from .types import Encoder, RawMessage

if TYPE_CHECKING:
    from eventiq import Consumer, Service

TOPIC_PATTERN = re.compile(r"{\w+}")


class Broker(Generic[RawMessage], LoggerMixin, ABC):
    """Base broker class
    :param description: Broker (Server) Description
    :param encoder: Encoder (Serializer) class
    :param middlewares: Optional list of middlewares
    """

    protocol: str
    protocol_version: str | None = None

    Settings = BrokerSettings

    WILDCARD_ONE: str
    WILDCARD_MANY: str

    def __init__(
        self,
        *,
        description: str | None = None,
        encoder: Encoder | type[Encoder] | None = None,
        middlewares: list[Middleware] | None = None,
    ) -> None:

        if encoder is None:
            from .encoders import get_default_encoder

            encoder = get_default_encoder()
        elif isinstance(encoder, type):
            encoder = encoder()
        self.encoder = encoder

        self.description = description or type(self).__name__
        self.middlewares: list[Middleware] = middlewares or []
        self._lock = anyio.Lock()
        self._running = False

    def __repr__(self):
        return type(self).__name__

    @property
    def message_proxy_class(self) -> type[Message]:
        return Message

    def get_handler(
        self, service: Service, consumer: Consumer
    ) -> Callable[..., Coroutine[Any, Any, Any]]:
        async def handler(raw_message: RawMessage) -> None:
            exc: Exception | None = None
            result: Any = None
            msg = self.message_proxy_class(raw_message)
            s_token = _current_service.set(service)
            try:
                parsed = self.parse_incoming_message(raw_message)
                message = consumer.validate_message(parsed)
                message.set_raw(msg)
            except (DecodeError, ValidationError) as e:
                self.logger.exception("Message parsing error", exc_info=e)
                msg.fail()
                await self.ack(service, consumer, msg)
                return
            m_token = _current_message.set(message)
            try:
                await self.dispatch_before(
                    "process_message", service, consumer, message
                )
            except Skip:
                self.logger.info(f"Skipped message {message.id}")
                await self.dispatch_after("skip_message", service, consumer, message)
                await self.ack(service, consumer, msg)
                return
            try:
                self.logger.info(
                    f"Running consumer {consumer.name} with message {message.id}"
                )
                async with anyio.move_on_after(consumer.timeout) as scope:
                    result = await consumer.process(message)

                    if consumer.forward_response and result is not None:
                        await self.publish(
                            CloudEvent(
                                type=consumer.forward_response.as_type,
                                topic=consumer.forward_response.topic,
                                data=result,
                                source=service.name,
                            )
                        )
                if scope.cancel_called:
                    exc = asyncio.TimeoutError()
            except Exception as e:
                self.logger.warning(
                    f"Exception in {consumer.name} <{type(e).__name__}> {e}"
                )
                exc = e
            finally:
                self.logger.info(f"Finished running consumer {consumer.name}")
                async with anyio.move_on_after(10, shield=True):
                    await self.dispatch_after(
                        "process_message", service, consumer, message, result, exc
                    )
                    if exc is None or isinstance(exc, Fail) or msg.failed:
                        await self.ack(service, consumer, msg)
                    else:
                        await self.nack(service, consumer, msg)
                _current_service.reset(s_token)
                _current_message.reset(m_token)

        return handler

    async def ack(self, service: Service, consumer: Consumer, message: Message) -> None:
        await self.dispatch_before("ack", service, consumer, message)
        await self._ack(message)
        await self.dispatch_after("ack", service, consumer, message)

    async def nack(
        self,
        service: Service,
        consumer: Consumer,
        message: Message,
    ) -> None:
        await self.dispatch_after(
            "nack",
            service,
            consumer,
            message,
        )
        await self._nack(message)
        await self.dispatch_after(
            "nack",
            service,
            consumer,
            message,
        )

    async def connect(self, **kwargs) -> None:
        async with self._lock:
            if not self._running:
                await self.dispatch_before("broker_connect")
                await self._connect()
                self._running = True
                await self.dispatch_after("broker_connect")

    async def disconnect(self) -> None:
        async with self._lock:
            if self._running:
                await self.dispatch_before("broker_disconnect")
                self._running = False
                await self._disconnect()
                await self.dispatch_after("broker_disconnect")

    async def publish(self, message: CloudEvent, **kwargs: Any) -> None:
        """
        :param message: Cloud event object to send
        :param kwargs: Additional params passed to broker._publish
        :rtype: None
        """
        await self.dispatch_before("publish", message)
        await self._publish(message, **kwargs)
        await self.dispatch_after("publish", message)

    async def start_consumer(self, service: Service, consumer: Consumer):
        await self.dispatch_before("consumer_start", service, consumer)
        await self._start_consumer(service, consumer)
        await self.dispatch_after("consumer_start", service, consumer)

    def add_middleware(self, middleware: Middleware | type[Middleware]) -> None:
        if isinstance(middleware, type):
            middleware = middleware()
        self.middlewares.append(middleware)

    def add_middlewares(self, middlewares: list[Middleware | type[Middleware]]) -> None:
        for m in middlewares:
            self.add_middleware(m)

    async def _dispatch(self, full_event: str, *args, **kwargs) -> None:
        for m in self.middlewares:
            try:
                await getattr(m, full_event)(self, *args, **kwargs)
            except Skip:
                raise
            except Exception as e:
                self.logger.exception("Unhandled middleware exception", exc_info=e)

    async def dispatch_before(self, event: str, *args, **kwargs) -> None:
        await self._dispatch(f"before_{event}", *args, **kwargs)

    async def dispatch_after(self, event: str, *args, **kwargs) -> None:
        await self._dispatch(f"after_{event}", *args, **kwargs)

    @classmethod
    def from_settings(cls, settings: BrokerSettings, **kwargs: Any) -> Broker:
        return cls(**settings.dict(), **kwargs)

    @classmethod
    def _from_env(cls, **kwargs) -> Broker:
        return cls.from_settings(cls.Settings(**kwargs))

    @classmethod
    def from_env(
        cls,
        **kwargs,
    ) -> Broker:
        if cls == Broker:
            type_name = os.getenv("BROKER_CLASS", "eventiq.backends.stub:StubBroker")
            broker_type = import_from_string(type_name)
        else:
            broker_type = cls
        return broker_type._from_env(**kwargs)

    @abstractmethod
    def parse_incoming_message(self, message: RawMessage) -> Any:
        raise NotImplementedError

    @property
    @abstractmethod
    def is_connected(self) -> bool:
        """Return broker connection status"""
        raise NotImplementedError

    @abstractmethod
    async def _publish(self, message: CloudEvent, **kwargs) -> None:
        raise NotImplementedError

    @abstractmethod
    async def _connect(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def _disconnect(self) -> None:
        raise NotImplementedError

    @abstractmethod
    async def _start_consumer(self, service: Service, consumer: Consumer) -> None:
        raise NotImplementedError

    async def _ack(self, message: Message) -> None:
        """Empty default implementation for backends that do not support explicit ack"""

    async def _nack(self, message: Message) -> None:
        """Same as for ._ack()"""

    def format_topic(self, topic: str) -> str:
        result = []
        for k in topic.split("."):
            if re.fullmatch(TOPIC_PATTERN, k):
                result.append(self.WILDCARD_ONE)
            elif k in {"*", ">"}:
                result.append(self.WILDCARD_MANY)
            else:
                result.append(k)
        return ".".join(filter(None, result))



import json
from typing import Any

from pydantic.json import pydantic_encoder

from ..exceptions import DecodeError, EncodeError


class JsonEncoder:
    """
    Default encoder, serializes CloudEvent models to json objects.
    """

    CONTENT_TYPE = "application/json"

    @staticmethod
    def encode(data: Any) -> bytes:
        try:
            return json.dumps(data, default=pydantic_encoder).encode("utf-8")
        except TypeError as e:
            raise EncodeError from e

    @staticmethod
    def decode(data: bytes) -> Any:
        try:
            return json.loads(data)
        except json.JSONDecodeError as e:
            raise DecodeError from e



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Callable, Sequence

import anyio

from .asyncapi.models import PublishInfo
from .asyncapi.registry import PUBLISH_REGISTRY
from .consumer import Consumer, ConsumerGroup, ForwardResponse
from .logger import LoggerMixin
from .models import CloudEvent
from .settings import DEFAULT_TIMEOUT, ServiceSettings
from .utils import generate_instance_id

if TYPE_CHECKING:
    from eventiq import Broker

    from .middlewares.retries import RetryStrategy
    from .types import TagMeta, Tags


class Service(LoggerMixin):
    """Logical group of consumers. Provides group (queue) name and handles versioning"""

    def __init__(
        self,
        name: str,
        broker: Broker,
        title: str | None = None,
        version: str = "0.1.0",
        description: str = "",
        tags_metadata: list[TagMeta] | None = None,
        instance_id_generator: Callable[[], str] | None = None,
        base_event_class: type[CloudEvent] = CloudEvent,
        publish_info: Sequence[PublishInfo] = (),
        **context: Any,
    ):
        self.broker = broker
        self.name = name
        self.title = title or name.title()
        self.version = version
        self.description = description
        self.tags_metadata = tags_metadata or []
        self.id = (instance_id_generator or generate_instance_id)()
        self.consumer_group = ConsumerGroup()
        self.context = context
        self.base_event_class = base_event_class
        for p in publish_info:
            PUBLISH_REGISTRY[p.event_type.__name__] = p

    def subscribe(
        self,
        topic: str,
        *,
        name: str | None = None,
        timeout: int = DEFAULT_TIMEOUT,
        dynamic: bool = False,
        forward_response: ForwardResponse | None = None,
        tags: Tags = None,
        retry_strategy: RetryStrategy | None = None,
        store_results: bool = False,
        parameters: dict[str, Any] | None = None,
        **options: Any,
    ):
        return self.consumer_group.subscribe(
            topic=topic,
            name=name,
            timeout=timeout,
            dynamic=dynamic,
            forward_response=forward_response,
            tags=tags,
            retry_strategy=retry_strategy,
            store_results=store_results,
            parameters=parameters,
            **options,
        )

    def add_consumer(self, consumer: Consumer):
        self.consumer_group.add_consumer(consumer)

    def add_consumer_group(self, consumer_group: ConsumerGroup) -> None:
        self.consumer_group.add_consumer_group(consumer_group)

    async def send(
        self,
        topic: str,
        type_: type[CloudEvent] | str = "CloudEvent",
        data: Any | None = None,
        **kwargs,
    ):
        if isinstance(type_, str):
            cls = self.base_event_class
            type_name = type_
        else:
            cls = type_
            type_name = type_.__name__

        message: CloudEvent = cls(
            content_type=self.broker.encoder.CONTENT_TYPE,
            type=type_name,
            topic=topic,
            data=data,
            source=self.name,
            **kwargs,
        )
        return await self.broker.publish(message)

    @property
    def consumers(self):
        return self.consumer_group.consumers

    async def publish(self, message: CloudEvent, **kwargs):
        if not message.source:
            message.set_source(self.name)
        return await self.broker.publish(message, **kwargs)

    async def start(self):
        self.logger.info(f"Starting service {self.name}")
        await self.broker.connect()
        await self.broker.dispatch_before("service_start", self)

        async with anyio.create_task_group() as tg:
            for consumer in self.consumers.values():
                self.logger.info(f"Starting consumer {consumer.name}")
                tg.start_soon(self.broker.start_consumer, self, consumer)
            await self.broker.dispatch_after("service_start", self)

    async def stop(self, *args, **kwargs):
        await self.broker.dispatch_before("service_stop", self)
        await self.broker.disconnect()
        await self.broker.dispatch_after("service_stop", self)

    async def run(self):
        from .runner import ServiceRunner

        runner = ServiceRunner([self])
        await runner.run()

    @classmethod
    def from_settings(cls, settings: ServiceSettings, **kwargs: Any) -> Service:
        return cls(**settings.dict(), **kwargs)

    @classmethod
    def from_env(cls, **kwargs: Any) -> Service:
        return cls.from_settings(ServiceSettings(), **kwargs)



from typing import Optional

from pydantic import BaseSettings, Field


class ConfigSettings(BaseSettings):
    ENV: Optional[str] = Field(None, env="ENV")
    config_file: str = Field("./config.yaml", env="CONFIG_FILE_PATH")



from .debug import DebugMiddleware
from .error import ErrorHandlerMiddleware
from .healthcheck import HealthCheckMiddleware
from .prometheus import PrometheusMiddleware
from .retries import RetryMiddleware

__all__ = [
    "DebugMiddleware",
    "ErrorHandlerMiddleware",
    "PrometheusMiddleware",
    "HealthCheckMiddleware",
    "RetryMiddleware",
]



from __future__ import annotations

import os

import yaml

from eventiq import Service, ServiceRunner

from .models import AppConfig


def create_app(config_file: str, section: str | None = None) -> ServiceRunner:

    with open(config_file) as f:
        data = f.read()
    formatted = data.format(**os.environ)
    loaded = yaml.safe_load(formatted)
    parsed = AppConfig.parse_obj(loaded[section] if section else loaded)
    _brokers = {}
    services = []
    for service_config in parsed.services:
        if service_config.broker not in _brokers:
            _brokers[service_config.broker] = parsed.brokers[
                service_config.broker
            ].build()
        service = Service(
            broker=_brokers[service_config.broker],
            **service_config.dict(exclude={"consumers", "broker"}),
        )
        for consumer_config in service_config.consumers:
            consumer = consumer_config.build()
            service.add_consumer(consumer)
        services.append(service)

    return ServiceRunner(services)


def create_app_from_env():
    from .settings import ConfigSettings

    settings = ConfigSettings()
    return create_app(settings.config_file, section=settings.ENV)



"""
To be done: allow auto-importing/registering third party plugins (from external packages) automatically on service startup
"""

from __future__ import annotations

from typing import Generic, TypeVar

from .broker import Broker
from .logger import LoggerMixin
from .service import Service

BrokerT = TypeVar("BrokerT", bound=Broker)


class ServicePlugin(LoggerMixin):
    def __init__(self, service: Service) -> None:
        self.service = service


class BrokerPlugin(Generic[BrokerT], LoggerMixin):
    def __init__(self, broker: BrokerT) -> None:
        self.broker = broker



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any

from eventiq.middleware import Middleware
from eventiq.utils import to_async

if TYPE_CHECKING:
    from eventiq import Broker, CloudEvent, Consumer, Service


class ErrorHandlerMiddleware(Middleware):
    def __init__(self, errors: type[Exception] | tuple[type[Exception]], callback):
        if not asyncio.iscoroutinefunction(callback):
            callback = to_async(callback)
        self.cb = callback
        self.exc = errors

    async def after_process_message(
        self,
        broker: Broker,
        service: Service,
        consumer: Consumer,
        message: CloudEvent,
        result: Any | None = None,
        exc: Exception | None = None,
    ):
        if exc and isinstance(exc, self.exc):
            await self.cb(message, exc)



from eventiq.types import Encoder


def get_default_encoder() -> Encoder:
    try:
        from eventiq.encoders.orjson import OrjsonEncoder

        return OrjsonEncoder()

    except ImportError:
        from eventiq.encoders.json import JsonEncoder

        return JsonEncoder()



from __future__ import annotations

from aio_pika import ExchangeType

from eventiq.backends.rabbitmq.broker import RabbitmqBroker
from eventiq.middleware import Middleware


class DeadLetterQueueMiddleware(Middleware):
    def __init__(self, dlx_name: str = "dlx") -> None:
        self.dlx_name = dlx_name
        self._dlx_exchange = None

    async def after_broker_connect(self, broker: RabbitmqBroker) -> None:  # type: ignore[override]
        # FIXME: implement properly
        assert isinstance(
            broker, RabbitmqBroker
        ), f"RabbitmqBroker instance expected, got {type(broker).__name__}"
        channel = await broker.connection.channel()
        self._dlx_exchange = await channel.declare_exchange(
            name=self.dlx_name, type=ExchangeType.HEADERS, auto_delete=True
        )



from __future__ import annotations

from contextvars import ContextVar
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from .models import CloudEvent
    from .service import Service


_current_message: ContextVar[CloudEvent | None] = ContextVar(
    "_current_message", default=None
)
_current_service: ContextVar[Service] = ContextVar("_current_service")


def get_current_message():
    global _current_message
    return _current_message.get()


def get_current_service() -> Service:
    global _current_service
    return _current_service.get()



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from nats.js.errors import KeyNotFoundError
from nats.js.kv import KeyValue

from eventiq.backends.nats.broker import JetStreamBroker
from eventiq.middleware import Middleware
from eventiq.plugins import BrokerPlugin
from eventiq.types import ID, ResultBackend
from eventiq.utils import retry

if TYPE_CHECKING:
    from eventiq import Broker, CloudEvent, Consumer, Service


class _NatsJetStreamResultMiddleware(Middleware):
    def __init__(self, result_backend: JetStreamResultBackend):
        self.result_backend = result_backend

    async def after_service_start(self, broker: Broker, service: Service):
        self.result_backend.buckets[service.name] = await broker.js.create_key_value(  # type: ignore[attr-defined]
            bucket=service.name, **self.result_backend.options
        )

    @retry(max_retries=3, backoff=10)
    async def after_process_message(
        self,
        broker: Broker,
        service: Service,
        consumer: Consumer,
        message: CloudEvent,
        result: Any | None = None,
        exc: Exception | None = None,
    ) -> None:
        """Store message result in JetStream K/V Store"""
        if not consumer.store_results:
            return

        kv = self.result_backend.buckets.get(service.name)

        if kv is None:
            self.logger.warning(f"Bucket {service.name} not found")
            return

        if exc is None:
            data = broker.encoder.encode(result)
        elif exc and self.result_backend.store_exceptions:
            data = broker.encoder.encode(
                {"type": type(exc).__name__, "detail": str(exc)}
            )
        else:
            return
        await kv.put(str(message.id), data)


class JetStreamResultBackend(BrokerPlugin[JetStreamBroker], ResultBackend):
    def __init__(
        self, broker: JetStreamBroker, store_exceptions: bool = False, **options: Any
    ):
        super().__init__(broker)
        self.store_exceptions = store_exceptions
        self.options = options
        self.broker.add_middleware(_NatsJetStreamResultMiddleware(self))
        self.buckets: dict[str, KeyValue] = {}

    @retry(max_retries=3)
    async def _get(self, kv: KeyValue, key: ID) -> Any:
        return await kv.get(str(key))

    async def get_result(self, service: str, message_id: ID) -> Any:
        kv = self.buckets.get(service)
        if kv is None:
            self.logger.warning(f"Bucket {service} not found")
            return

        try:
            data = await self._get(kv, message_id)
            return self.broker.encoder.decode(data.value)
        except KeyNotFoundError:
            self.logger.warning(f"Key {message_id} not found")
            return



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from eventiq import Middleware, Service
from eventiq.plugins import BrokerPlugin
from eventiq.types import ID, ResultBackend

from ...utils import retry
from .broker import RedisBroker

if TYPE_CHECKING:
    from eventiq import CloudEvent, Consumer


class _RedisResultMiddleware(Middleware):
    def __init__(self, store_exceptions: bool, ttl: int):
        self.store_exceptions = store_exceptions
        self.ttl = ttl

    @retry(max_retries=3, backoff=10)
    async def after_process_message(
        self,
        broker: RedisBroker,
        service: Service,
        consumer: Consumer,
        message: CloudEvent,
        result: Any | None = None,
        exc: Exception | None = None,
    ) -> None:
        if not consumer.store_results:
            return

        if exc is None:
            data = broker.encoder.encode(result)

        elif exc and self.store_exceptions:
            data = broker.encoder.encode(
                {"type": type(exc).__name__, "detail": str(exc)}
            )
        else:
            return
        await broker.redis.set(
            f"{service.name}:{consumer.name}:{message.id}", data, ex=self.ttl
        )


class RedisResultBackend(BrokerPlugin[RedisBroker], ResultBackend):
    def __init__(
        self, broker: RedisBroker, store_exceptions: bool = False, ttl: int = 3600
    ):
        super().__init__(broker)
        broker.add_middleware(_RedisResultMiddleware(store_exceptions, ttl))

    async def get_result(self, service: str, message_id: ID):
        res = await self.broker.redis.get(f"{service}:{message_id}")
        if res:
            return self.broker.encoder.decode(res)



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import aiokafka
import anyio

from eventiq.broker import Broker
from eventiq.exceptions import BrokerError

from .settings import KafkaSettings

if TYPE_CHECKING:
    from eventiq import CloudEvent, Consumer, Service


class KafkaBroker(Broker[aiokafka.ConsumerRecord]):
    """
    Kafka backend
    :param bootstrap_servers: url or list of kafka servers
    :param publisher_options: extra options for AIOKafkaProducer
    :param consumer_options: extra options (defaults) for AIOKafkaConsumer
    :param kwargs: Broker base class parameters
    """

    WILDCARD_MANY = "*"
    WILDCARD_ONE = r"\w+"

    Settings = KafkaSettings
    protocol = "kafka"

    def __init__(
        self,
        *,
        bootstrap_servers: str | list[str],
        publisher_options: dict[str, Any] | None = None,
        consumer_options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:

        super().__init__(**kwargs)
        self.bootstrap_servers = bootstrap_servers
        self._publisher_options = publisher_options or {}
        self._consumer_options = consumer_options or {}
        self._publisher = None

    def parse_incoming_message(self, message: aiokafka.ConsumerRecord) -> Any:
        return self.encoder.decode(message.value)

    @property
    def is_connected(self) -> bool:
        return True

    async def _start_consumer(self, service: Service, consumer: Consumer) -> None:
        handler = self.get_handler(service, consumer)
        subscriber = aiokafka.AIOKafkaConsumer(
            group_id=f"{service.name}:{consumer.name}",
            bootstrap_servers=self.bootstrap_servers,
            enable_auto_commit=False,
            **consumer.options.get("kafka_consumer_options", self._consumer_options),
        )
        await subscriber.start()
        subscriber.subscribe(pattern=self.format_topic(consumer.topic))
        try:
            while self._running:
                result = await subscriber.getmany(
                    timeout_ms=consumer.options.get("timeout_ms", 600)
                )

                for tp, messages in result.items():
                    if messages:
                        async with anyio.create_task_group() as tg:
                            for message in messages:
                                tg.start_soon(handler, message)
                        await subscriber.commit({tp: messages[-1].offset + 1})
        finally:
            if consumer.dynamic:
                subscriber.unsubscribe()

    async def _disconnect(self):
        if self._publisher:
            await self._publisher.stop()

    @property
    def publisher(self) -> aiokafka.AIOKafkaProducer:
        if self._publisher is None:
            raise BrokerError("Broker not connected")
        return self._publisher

    async def _connect(self):
        self._publisher = aiokafka.AIOKafkaProducer(
            bootstrap_servers=self.bootstrap_servers, **self._publisher_options
        )
        await self._publisher.start()

    async def _publish(
        self,
        message: CloudEvent,
        key: Any | None = None,
        partition: Any | None = None,
        headers: dict[str, str] | None = None,
        timestamp_ms: int | None = None,
        **kwargs: Any,
    ):
        data = self.encoder.encode(message.dict())
        timestamp_ms = timestamp_ms or int(message.time.timestamp() * 1000)
        key = key or getattr(message, "key", str(message.id))
        headers = headers or {}
        headers.setdefault("Content-Type", self.encoder.CONTENT_TYPE)
        await self.publisher.send(
            topic=message.topic,
            value=data,
            key=key,
            partition=partition,
            headers=headers,
            timestamp_ms=timestamp_ms,
        )



from __future__ import annotations

from eventiq.asyncapi.models import PublishInfo
from eventiq.models import CloudEvent

PUBLISH_REGISTRY: dict[str, PublishInfo] = {}


def publishes(topic: str | None = None, **kwargs):
    def wrapper(cls: type[CloudEvent]) -> type[CloudEvent]:
        PUBLISH_REGISTRY[cls.__name__] = PublishInfo(
            topic=topic, event_type=cls, kwargs=kwargs
        )
        return cls

    return wrapper



from __future__ import annotations

from datetime import timedelta
from typing import TYPE_CHECKING, Any, Callable, Mapping

from eventiq.exceptions import Fail, Retry
from eventiq.logger import LoggerMixin
from eventiq.middleware import Middleware

if TYPE_CHECKING:
    from eventiq import Broker, CloudEvent, Consumer, Service


class RetryStrategy(LoggerMixin):
    def __init__(
        self,
        backoff: int = 2,
        throws: tuple[type[Exception], ...] = (),
    ):
        self.backoff = backoff
        if Fail not in throws:
            throws = (*throws, Fail)
        self.throws = throws

    def set_delay(self, message: CloudEvent, exc: Exception):
        delay = getattr(exc, "delay", None) or self.backoff * message.age.seconds
        message.raw.delay = delay
        self.logger.info("Will retry message in %d seconds.", delay)

    def fail(self, message: CloudEvent, exc: Exception):
        message.fail()
        self.logger.error(f"Retry limit exceeded for message {message.id}.")
        self.logger.exception("Original exception:", exc_info=exc)

    def maybe_retry(self, message: CloudEvent, exc: Exception):
        if self.throws and isinstance(exc, self.throws):
            message.fail()
        else:
            self.set_delay(message, exc)


class MaxAge(RetryStrategy):
    def __init__(
        self, max_age: timedelta | dict[str, Any] = timedelta(seconds=60), **extra
    ):
        super().__init__(**extra)
        if isinstance(max_age, Mapping):
            max_age = timedelta(**max_age)
        self.max_age: timedelta = max_age

    def maybe_retry(self, message: CloudEvent, exc: Exception):
        if message.age <= self.max_age:
            super().maybe_retry(message, exc)
        else:
            self.fail(message, exc)


class MaxRetries(RetryStrategy):
    def __init__(self, max_retries: int = 3, **extra):
        super().__init__(**extra)
        self.max_retries = max_retries

    def maybe_retry(self, message: CloudEvent, exc: Exception):
        retries = message.raw.num_delivered
        if retries is None:
            self.logger.warning(
                "Retries property not found in message, backing off to message.age.seconds"
            )
            retries = message.age.seconds
        if retries <= self.max_retries:
            super().maybe_retry(message, exc)
        else:
            self.fail(message, exc)


class RetryWhen(RetryStrategy):
    def __init__(self, retry_when: Callable[[CloudEvent, Exception], bool], **extra):
        super().__init__(**extra)
        self.retry_when = retry_when

    def maybe_retry(
        self,
        message: CloudEvent,
        exc: Exception,
    ):
        if self.retry_when(message, exc):
            super().maybe_retry(message, exc)
        else:
            self.fail(message, exc)


class RetryMiddleware(Middleware):
    """
    Retry Message Middleware.
    Supported retry strategies:
    - `MaxAge` (default) - retry with exponential backoff up to max_age
    - `MaxRetries` - retry up to N times (currently supported only by nats)
    - `RetryWhen` - provide custom callable to determine weather message should be retried
    """

    def __init__(self, default_retry_strategy: RetryStrategy | None = None):
        self.default_retry_strategy = default_retry_strategy or MaxAge(
            max_age=timedelta(hours=1)
        )

    async def after_process_message(
        self,
        broker: Broker,
        service: Service,
        consumer: Consumer,
        message: CloudEvent,
        result: Any | None = None,
        exc: Exception | None = None,
    ):

        if exc is None:
            return

        if isinstance(exc, Retry):
            message.raw.delay = exc.delay
            return
        retry_strategy = consumer.retry_strategy or self.default_retry_strategy

        if retry_strategy is None:
            return

        retry_strategy.maybe_retry(message, exc)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ContextManager, Iterable

from opentelemetry import trace
from opentelemetry.propagate import extract, inject
from opentelemetry.propagators.textmap import Getter, Setter
from opentelemetry.semconv.trace import SpanAttributes
from opentelemetry.trace import SpanKind, StatusCode

from .._version import __version__
from ..middleware import Middleware
from ..models import CloudEvent

if TYPE_CHECKING:
    from opentelemetry.sdk.trace import Span, TracerProvider

    from eventiq import Broker, Consumer, Service
    from eventiq.types import ID


class EventiqGetter(Getter[CloudEvent]):
    def get(self, carrier: CloudEvent, key: str) -> list[str] | None:
        val = carrier.tracecontext.get(key, None)
        if val is None:
            return None
        if isinstance(val, Iterable) and not isinstance(val, str):
            return list(val)
        return [val]

    def keys(self, carrier: CloudEvent) -> list[str]:
        return list(carrier.tracecontext.keys())


class EventiqSetter(Setter[CloudEvent]):
    def set(self, carrier: CloudEvent, key: str, value: str) -> None:
        carrier.tracecontext[key] = value


eventiq_getter = EventiqGetter()
eventiq_setter = EventiqSetter()


class OpenTelemetryMiddleware(Middleware):
    def __init__(self, provider: TracerProvider | None = None):
        if provider is None:
            provider = trace.get_tracer_provider()

        self.tracer = provider.get_tracer("eventiq", __version__)
        self.process_span_registry: dict[
            tuple[str, str, ID], tuple[Span, ContextManager[Span]]
        ] = {}
        self.publish_span_registry: dict[ID, tuple[Span, ContextManager[Span]]] = {}

    @staticmethod
    def _get_span_attributes(message: CloudEvent):
        return {
            SpanAttributes.CLOUDEVENTS_EVENT_ID: str(message.id),
            SpanAttributes.CLOUDEVENTS_EVENT_SOURCE: message.source or "(anonymous)",
            SpanAttributes.CLOUDEVENTS_EVENT_TYPE: message.type or "CloudEvent",
            SpanAttributes.CLOUDEVENTS_EVENT_SUBJECT: message.topic,
            **message.extra_span_attributes,
        }

    async def before_process_message(
        self, broker: Broker, service: Service, consumer: Consumer, message: CloudEvent
    ) -> None:
        trace_ctx = extract(message, getter=eventiq_getter)

        span = self.tracer.start_span(
            name=f"{consumer.name} receive",
            kind=SpanKind.CONSUMER,
            context=trace_ctx,
            attributes=self._get_span_attributes(message),
        )
        activation = trace.use_span(span, end_on_exit=True)
        activation.__enter__()
        self.process_span_registry[(service.name, consumer.name, message.id)] = (
            span,
            activation,
        )

    async def after_process_message(
        self,
        broker: Broker,
        service: Service,
        consumer: Consumer,
        message: CloudEvent,
        result: Any | None = None,
        exc: Exception | None = None,
    ) -> None:
        key = (service.name, consumer.name, message.id)
        span, activation = self.process_span_registry.pop(key, (None, None))
        if span is None or activation is None:
            self.logger.warning("No active span was found")
            return

        if span.is_recording():
            if exc:
                span.set_status(
                    status=StatusCode.ERROR, description=str(exc) or type(exc).__name__
                )
            else:
                span.set_status(StatusCode.OK)
        activation.__exit__(None, None, None)

    async def before_publish(
        self, broker: Broker, message: CloudEvent, **kwargs
    ) -> None:
        source = message.source or "(anonymous)"

        span = self.tracer.start_span(
            f"{source} publish",
            kind=SpanKind.PRODUCER,
            attributes=self._get_span_attributes(message),
        )
        activation = trace.use_span(span, end_on_exit=True)
        activation.__enter__()
        self.publish_span_registry[message.id] = (span, activation)
        inject(message, setter=eventiq_setter)

    async def after_publish(
        self, broker: Broker, message: CloudEvent, **kwargs
    ) -> None:
        span, activation = self.publish_span_registry.pop(message.id, (None, None))
        if span and span.is_recording():
            span.set_status(StatusCode.OK)
        if activation is not None:
            activation.__exit__(None, None, None)



import asyncio
import functools
import time

from eventiq.middleware import Middleware


def _logged_coro(func):
    @functools.wraps(func)
    async def wrapper(self, *args, **kwargs):
        self.logger.debug(f"Running func {func.__name__} with {args}, {kwargs}")
        start = time.perf_counter()
        try:
            return await func(self, *args, **kwargs)
        finally:
            delta = time.perf_counter() - start
            self.logger.debug("Completed after %.02fms.", delta * 1000)

    return wrapper


class DebugMeta(type):
    def __new__(mcs, name, bases, namespace):
        for k, v in namespace.items():
            if asyncio.iscoroutinefunction(v):
                namespace[k] = _logged_coro(v)
        return type.__new__(mcs, name, bases, namespace)


class DebugMiddleware(Middleware, metaclass=DebugMeta):
    """Log every event"""



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from eventiq.middleware import Middleware
from eventiq.types import ID
from eventiq.utils import current_millis

if TYPE_CHECKING:
    from prometheus_client.registry import CollectorRegistry

    from eventiq import Broker, CloudEvent, Consumer, Message, Service


DEFAULT_BUCKETS = (
    5,
    10,
    25,
    50,
    75,
    100,
    250,
    500,
    750,
    1000,
    2500,
    5000,
    7500,
    10000,
    30000,
    60000,
    600000,
    900000,
    float("inf"),
)


class PrometheusMiddleware(Middleware):
    def __init__(
        self,
        run_server: bool = False,
        registry: CollectorRegistry | None = None,
        buckets: tuple[float] | None = None,
        server_host: str = "0.0.0.0",  # nosec
        server_port: int = 8888,
        prefix: str = "",
    ):
        from prometheus_client import REGISTRY, Counter, Gauge, Histogram

        self.run_server = run_server
        self.registry = registry or REGISTRY
        self.buckets = buckets or DEFAULT_BUCKETS
        self.server_host = server_host
        self.server_port = server_port
        self.message_start_times: dict[tuple[str, str, ID], int] = {}
        self.prefix = prefix
        self.in_progress = Gauge(
            self.format("messages_in_progress"),
            "Total number of messages being processed.",
            ["topic", "service", "consumer"],
            registry=self.registry,
        )
        self.total_messages = Counter(
            self.format("messages_total"),
            "Total number of messages processed.",
            ["topic", "service", "consumer"],
            registry=self.registry,
        )
        self.total_skipped_messages = Counter(
            self.format("messages_skipped_total"),
            "Total number of messages skipped processing.",
            registry=self.registry,
        )
        self.total_messages_published = Counter(
            self.format("messages_published_total"),
            "Total number of messages published",
            ["topic", "service"],
            registry=self.registry,
        )
        self.total_errored_messages = Counter(
            self.format("message_error_total"),
            "Total number of errored messages.",
            ["topic", "service", "consumer"],
            registry=self.registry,
        )
        self.total_failed_messages = Counter(
            self.format("message_failed_total"),
            "Total number of messages failed",
            ["topic", "service", "consumer"],
            registry=self.registry,
        )
        self.message_durations = Histogram(
            self.format("message_duration_ms"),
            "Time spend processing message",
            ["topic", "service", "consumer"],
            registry=self.registry,
            buckets=self.buckets,
        )

    def format(self, value: str) -> str:
        if self.prefix:
            return f"{self.prefix}_{value}"
        return value

    async def before_process_message(
        self, broker: Broker, service: Service, consumer: Consumer, message: CloudEvent
    ):
        labels = (consumer.topic, service.name, consumer.name)
        self.in_progress.labels(*labels).inc()
        self.message_start_times[
            (service.name, consumer.name, message.id)
        ] = current_millis()

    async def after_process_message(
        self,
        broker: Broker,
        service: Service,
        consumer: Consumer,
        message: CloudEvent,
        result: Any | None = None,
        exc: Exception | None = None,
    ) -> None:
        labels = (consumer.topic, service.name, consumer.name)
        self.in_progress.labels(*labels).dec()
        self.total_messages.labels(*labels).inc()
        if exc:
            self.total_errored_messages.labels(*labels).inc()

        message_start_time = self.message_start_times.pop(
            (service.name, consumer.name, message.id), current_millis()
        )
        message_duration = current_millis() - message_start_time
        self.message_durations.labels(*labels).observe(message_duration)

    async def after_skip_message(
        self, broker: Broker, service: Service, consumer: Consumer, message: CloudEvent
    ) -> None:
        labels = (consumer.topic, service.name, consumer.name)
        self.total_skipped_messages.labels(*labels).inc()

    async def after_publish(self, broker: Broker, message: CloudEvent, **kwargs):
        self.total_messages_published.labels(message.topic, message.source).inc()

    async def after_nack(
        self, broker: Broker, service: Service, consumer: Consumer, message: Message
    ):
        labels = (consumer.topic, service.name, consumer.name)
        self.total_errored_messages.labels(*labels).inc()

    async def after_ack(
        self,
        broker: Broker,
        service: Service,
        consumer: Consumer,
        message: Message,
    ) -> None:
        if message.failed:
            labels = (consumer.topic, service.name, consumer.name)
            self.total_failed_messages.labels(*labels).inc()

    async def after_broker_connect(self, broker: Broker):
        if self.run_server:
            from prometheus_client import start_http_server

            start_http_server(
                self.server_port, self.server_host, registry=self.registry
            )



from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Extra

from eventiq.consumer import FnConsumer, GenericConsumer
from eventiq.types import TagMeta
from eventiq.utils.imports import ImportedType


def resolve_nested(v: Any):
    if isinstance(v, TypedModel):
        return v.build()
    elif isinstance(v, list):
        v = [resolve_nested(i) for i in v]
    if isinstance(v, dict):
        if "type" in v:
            return TypedModel.parse_obj(v).build()
        else:
            for key, value in v.items():
                v[key] = resolve_nested(value)
    return v


class TypedModel(BaseModel):
    type: ImportedType

    def build(self):
        kwargs = self.dict(exclude={"type"}, exclude_none=True)
        for k, v in kwargs.items():
            kwargs[k] = resolve_nested(v)
        return self.type(**kwargs)

    class Config:
        extra = Extra.allow


class BrokerConfig(TypedModel):
    encoder: TypedModel
    middlewares: List[TypedModel]
    context: Dict[str, Union[TypedModel, Any]] = {}


class ConsumerConfig(TypedModel):
    topic: str
    name: Optional[str]
    timeout: int = 120
    dynamic: bool = False

    def build(self):
        if callable(self.type) and not (
            isinstance(self.type, type) and issubclass(self.type, GenericConsumer)
        ):
            self.__dict__["fn"] = self.type
            self.type = FnConsumer
        return super().build()


class ServiceConfig(BaseModel):
    name: str
    broker: str = "default"
    title: Optional[str]
    version: str = "0.1.0"
    description: str = ""
    tags_metadata: List[TagMeta] = []
    instance_id_generator: Optional[ImportedType]
    consumers: List[ConsumerConfig]


class AppConfig(BaseModel):
    brokers: Dict[str, BrokerConfig]
    services: List[ServiceConfig]



from __future__ import annotations

import asyncio

from fastapi import Depends, FastAPI, Request
from fastapi.responses import JSONResponse

from ..asyncapi.generator import get_async_api_spec
from ..asyncapi.models import AsyncAPI
from ..plugins import Service, ServicePlugin


class FastAPIServicePlugin(ServicePlugin):
    """Integration with fastapi, allows running service alongside fastapi app"""

    def __init__(
        self,
        service: Service,
    ):
        super().__init__(service)
        self._task: asyncio.Task | None = None

    def configure_app(
        self,
        app: FastAPI,
        run_service: bool = True,
        async_api_url: str | None = None,
        healthcheck_url: str | None = None,
    ) -> None:
        app._service = self.service
        if run_service:
            self._add_run_service_events(app)

        else:
            self._add_broker_connect_events(app)

        if async_api_url:
            self._add_asyncapi_url(app, async_api_url)

        if healthcheck_url:
            self._add_healthcheck(app, healthcheck_url)

    def _add_run_service_events(self, app):
        @app.on_event("startup")
        async def create_service_task():
            self._task = asyncio.create_task(self.service.start())

        @app.on_event("shutdown")
        async def stop_service_task():
            if self._task:
                self._task.cancel()
                await asyncio.wait_for(self._task, timeout=15)
                await self.service.stop()

    def _add_broker_connect_events(self, app):
        @app.on_event("startup")
        async def connect_broker():
            await self.service.broker.connect()

        @app.on_event("shutdown")
        async def disconnect_broker():
            await self.service.broker.disconnect()

    def _add_asyncapi_url(self, app, async_api_url):
        @app.get(async_api_url, response_model=AsyncAPI)
        def get_service_asyncapi():
            """Return service Async API specification"""
            return get_async_api_spec(self.service)

    def _add_healthcheck(self, app, healthcheck_url):
        @app.get(healthcheck_url, response_class=JSONResponse)
        def get_broker_connection_status():
            try:
                res = self.service.broker.is_connected
                if not res:
                    raise ValueError
                return JSONResponse({"status": "ok"})
            except Exception:
                return JSONResponse(
                    {"detail": "Service Unavailable", "status": 503},
                    status_code=503,
                )


def get_service(request: Request) -> Service:
    return request.app._service


def ProvideService():
    return Depends(get_service)



from typing import Any, Dict, List, Optional, Union

from pydantic import Field

from eventiq.settings import BrokerSettings


class KafkaSettings(BrokerSettings):
    bootstrap_servers: Union[str, List[str]] = Field(
        ..., env="BROKER_BOOTSTRAP_SERVERS"
    )
    publisher_options: Optional[Dict[str, Any]] = Field(
        None, env="BROKER_PUBLISHER_OPTIONS"
    )
    consumer_options: Optional[Dict[str, Any]] = Field(
        None, env="BROKER_CONSUMER_OPTIONS"
    )



import os
from typing import List, Optional, Union

from pydantic import BaseSettings, Field

from .imports import ImportedType
from .middleware import Middleware
from .types import Encoder

DEFAULT_TIMEOUT = int(os.getenv("DEFAULT_CONSUMER_TIMEOUT", "300"))


class BrokerSettings(BaseSettings):
    description: Optional[str] = None
    middlewares: Optional[List[Middleware]] = None
    encoder: Optional[Union[Encoder, ImportedType]] = Field(
        None, env="BROKER_ENCODER_CLASS"
    )


class ServiceSettings(BaseSettings):
    name: str
    title: Optional[str] = None
    version: str = "0.1.0"
    description: str = ""



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from gcloud.aio.pubsub import (
    PublisherClient,
    PubsubMessage,
    SubscriberClient,
    SubscriberMessage,
    subscribe,
)

from eventiq.broker import Broker
from eventiq.exceptions import BrokerError

from ...utils import retry
from .settings import PubSubSettings

if TYPE_CHECKING:
    from eventiq import CloudEvent, Consumer, Service


class PubSubBroker(Broker[SubscriberMessage]):
    """
    Google Cloud Pub/Sub broker implementation
    :param service_file: path to the service account (json) file
    :param kwargs: Broker base class parameters
    """

    Settings = PubSubSettings

    WILDCARD_ONE = "*"
    WILDCARD_MANY = "*"

    def __init__(
        self,
        *,
        service_file: str,
        **kwargs: Any,
    ) -> None:

        super().__init__(**kwargs)
        self.service_file = service_file
        self._client = None

    def parse_incoming_message(self, message: SubscriberMessage) -> Any:
        return self.encoder.decode(message.data)

    async def _disconnect(self) -> None:
        await self.client.close()

    async def _start_consumer(self, service: Service, consumer: Consumer) -> None:
        consumer_client = SubscriberClient(service_file=self.service_file)
        handler = self.get_handler(service, consumer)
        await subscribe(
            subscription=consumer.topic,
            handler=handler,
            subscriber_client=consumer_client,
            **consumer.options.get("subscribe_options", {}),
        )

    @property
    def client(self) -> PublisherClient:
        if self._client is None:
            raise BrokerError("Broker not connected")
        return self._client

    @retry(max_retries=3)
    async def _publish(
        self,
        message: CloudEvent,
        **kwargs: Any,
    ) -> None:
        ordering_key = kwargs.get("ordering_key", str(message.id))
        timeout = kwargs.get("timeout", 10)
        msg = PubsubMessage(
            data=self.encoder.encode(message.dict()),
            ordering_key=ordering_key,
            content_type=self.encoder.CONTENT_TYPE,
            **kwargs.get("headers", {}),
        )
        await self.client.publish(topic=message.topic, messages=[msg], timeout=timeout)

    async def _connect(self) -> None:
        self._client = PublisherClient(service_file=self.service_file)

    @property
    def is_connected(self) -> bool:
        return self.client.session._session.closed



from __future__ import annotations

from typing import Generic

from .types import RawMessage


class Message(Generic[RawMessage]):
    __slots__ = ("_message", "_failed", "_retry_delay")

    def __init__(self, message: RawMessage):
        self._message = message
        self._failed = False
        self._retry_delay: int | None = None

    def __getattr__(self, item):
        return getattr(self._message, item)

    def __str__(self):
        return str(self._message)

    @property
    def delay(self):
        return self._retry_delay

    @delay.setter
    def delay(self, value: int):
        if value < 0:
            raise ValueError("Cannot set negative delay")
        self._retry_delay = value

    @property
    def failed(self) -> bool:
        return self._failed

    def fail(self) -> None:
        self._failed = True

    @property
    def num_delivered(self) -> int | None:
        return None



from __future__ import annotations


class ConfigurationError(Exception):
    """Raised by framework when invalid configuration is supplied"""


class BrokerError(Exception):
    """Base Exception for broker related errors"""


class PublishError(BrokerError):
    """Raised when publishing a message fails"""


class EncoderError(Exception):
    """Base Encoder error"""


class EncodeError(EncoderError):
    """Error encoding message"""


class DecodeError(EncoderError):
    """Error decoding message"""


class Skip(Exception):
    """Raise exception to skip message without processing and/or retrying"""


class Fail(Exception):
    """Fail message without retrying"""


class Retry(Exception):
    """
    Utility exception for retrying message.
    RetryMiddleware must be added
    """

    def __init__(self, delay: int | None = None):
        self.delay = delay



from typing import Any

import orjson
from pydantic.json import pydantic_encoder

from ..exceptions import DecodeError, EncodeError


class OrjsonEncoder:
    """
    Json encoder which utilizes orjson library.
    """

    CONTENT_TYPE = "application/json"

    @staticmethod
    def encode(data: Any) -> bytes:
        try:
            return orjson.dumps(
                data,
                option=orjson.OPT_NON_STR_KEYS | orjson.OPT_SERIALIZE_NUMPY,
                default=pydantic_encoder,
            )
        except TypeError as e:
            raise EncodeError from e

    @staticmethod
    def decode(data: bytes) -> Any:
        try:
            return orjson.loads(data)
        except orjson.JSONDecodeError as e:
            raise DecodeError from e



from __future__ import annotations

import asyncio
import functools
import socket
import time
from datetime import datetime, timezone
from typing import Any, Awaitable, Callable, TypeVar

import anyio
from typing_extensions import ParamSpec

P = ParamSpec("P")
R = TypeVar("R", bound=Any)


def utc_now() -> datetime:
    return datetime.now(tz=timezone.utc)


def current_millis() -> int:
    return time.monotonic_ns() // 1000


def generate_instance_id() -> str:
    return socket.gethostname()


def to_async(func: Callable[P, R]) -> Callable[P, Awaitable[R]]:
    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> Awaitable[R]:
        return anyio.to_thread.run_sync(functools.partial(func, *args, **kwargs))

    return wrapper


def retry(max_retries: int = 5, backoff: int = 2):
    def _wrapper(
        func: Callable[P, R] | Callable[P, Awaitable[R]]
    ) -> Callable[P, R] | Callable[P, Awaitable[R]]:
        if asyncio.iscoroutinefunction(func):
            return _retry_async(func, max_retries, backoff)

        return _retry_sync(func, max_retries, backoff)

    return _wrapper


def _retry_async(
    func: Callable[P, Awaitable[R]] | Callable[P, R], max_retries: int, backoff: int
) -> Callable[P, R] | Callable[P, Awaitable[R]]:
    @functools.wraps(func)
    async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        exc = None
        for i in range(1, max_retries + 1):
            try:
                return await func(*args, **kwargs)
            except Exception as e:
                exc = e
                await asyncio.sleep(backoff**i)
        else:
            raise exc  # type: ignore

    return wrapper


def _retry_sync(func: Callable[P, R], max_retries: int, backoff: int) -> Callable[P, R]:
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        exc = None
        for i in range(1, max_retries + 1):
            try:
                return func(*args, **kwargs)
            except Exception as e:
                exc = e
                time.sleep(i**backoff)
        raise exc

    return wrapper



from typing import Any, Dict, List, Optional, Type

from pydantic import AnyUrl
from pydantic import BaseModel as _BaseModel
from pydantic import Field, validator

from eventiq.models import CloudEvent


class BaseModel(_BaseModel):
    class Config:
        allow_population_by_field_name = True


class ExtendableBaseModel(BaseModel):
    class Config:
        allow_population_by_field_name = True
        extra = "allow"


class Info(BaseModel):
    title: str
    version: str


class ExternalDocumentation(ExtendableBaseModel):
    description: Optional[str] = None
    url: AnyUrl


class Tag(ExtendableBaseModel):
    name: str
    description: Optional[str] = None
    externalDocs: Optional[ExternalDocumentation] = None

    class Config:
        extra = "allow"


class PublishInfo(BaseModel):
    event_type: Type[CloudEvent]
    topic: str
    kwargs: Dict[str, Any] = {}

    @validator("topic", always=True, pre=True, allow_reuse=True)
    def set_default_topic(cls, v, values):
        if v is None:
            v = values["event_type"].__fields__["topic"].get_default()
        return v

    @classmethod
    def s(cls, even_type: Type[CloudEvent], topic: Optional[str] = None, **kwargs: Any):
        return cls(event_type=even_type, topic=topic, kwargs=kwargs)


class Ref(BaseModel):
    ref: str = Field(..., alias="$ref")


class Message(BaseModel):
    payload: Ref
    content_type: str = Field(..., alias="contentType")
    description: Optional[str] = None
    tags: Optional[List[Tag]] = None


class Operation(BaseModel):
    operation_id: str = Field(..., alias="operationId")
    summary: Optional[str] = None
    message: Ref
    tags: Optional[List[Tag]] = None


class Parameter(BaseModel):
    description: Optional[str] = None
    param_schema: Dict[str, Any] = Field({"type": "string"}, alias="schema")
    location: Optional[str] = None


class ChannelItem(BaseModel):
    publish: Optional[Operation] = None
    subscribe: Optional[Operation] = None
    parameters: Dict[str, Parameter] = {}


class Server(BaseModel):
    protocol: str
    url: Optional[str] = None
    protocol_version: Optional[str] = Field(None, alias="protocolVersion")
    description: Optional[str] = None


class Components(BaseModel):
    schemas: Dict[str, Any] = {}
    messages: Dict[str, Any] = {}


class AsyncAPI(BaseModel):
    asyncapi: str = "2.5.0"
    info: Info
    servers: Dict[str, Server] = {}
    channels: Dict[str, ChannelItem] = {}
    default_content_type: str = Field("application/json", alias="defaultContentType")
    components: Components
    tags: Optional[List[Tag]] = None



from __future__ import annotations

from typing import TYPE_CHECKING, Any

import aio_pika

from eventiq.broker import Broker

from .settings import RabbitMQSettings

if TYPE_CHECKING:
    from eventiq import CloudEvent, Consumer, Service


class RabbitmqBroker(Broker[aio_pika.abc.AbstractIncomingMessage]):
    """
    RabbitMQ broker implementation, based on `aio_pika` library.
    :param url: rabbitmq connection string
    :param default_prefetch_count: default number of messages to prefetch (per queue)
    :param queue_options: additional queue options
    :param exchange_name: global exchange name
    :param connection_options: additional connection options passed to aio_pika.connect_robust
    :param kwargs: Broker base class parameters
    """

    Settings = RabbitMQSettings

    WILDCARD_ONE = "*"
    WILDCARD_MANY = "#"

    def __init__(
        self,
        *,
        url: str,
        default_prefetch_count: int = 10,
        queue_options: dict[str, Any] | None = None,
        exchange_name: str = "events",
        connection_options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:

        super().__init__(**kwargs)
        self.url = url
        self.default_prefetch_count = default_prefetch_count
        self.queue_options = queue_options or {}
        self.exchange_name = exchange_name
        self.connection_options = connection_options or {}
        self._connection = None
        self._exchange = None
        self._channels: list[aio_pika.abc.AbstractRobustChannel] = []

    @property
    def connection(self) -> aio_pika.RobustConnection:
        return self._connection

    @property
    def exchange(self) -> aio_pika.abc.AbstractRobustExchange:
        return self._exchange

    async def _connect(self) -> None:
        self._connection = await aio_pika.connect_robust(
            self.url, **self.connection_options
        )
        channel = await self.connection.channel()
        self._exchange = await channel.declare_exchange(
            name=self.exchange_name, type=aio_pika.ExchangeType.TOPIC, durable=True
        )

    async def _disconnect(self) -> None:
        for c in self._channels:
            await c.close()
        await self.connection.close()

    async def _start_consumer(self, service: Service, consumer: Consumer) -> None:
        """
        to route the messages to consumers
        :param service:
        :param consumer:
        :return:
        """
        channel = await self.connection.channel()
        await channel.set_qos(
            prefetch_count=consumer.options.get(
                "prefetch_count", self.default_prefetch_count
            )
        )
        options: dict[str, Any] = consumer.options.get(
            "queue_options", self.queue_options
        )
        is_durable = not consumer.dynamic
        options.setdefault("durable", is_durable)
        queue_name = f"{service.name}:{consumer.name}"
        queue = await channel.declare_queue(name=queue_name, **options)
        await queue.bind(self._exchange, routing_key=consumer.topic)
        handler = self.get_handler(service, consumer)
        await queue.consume(handler)
        self._channels.append(channel)

    async def _publish(self, message: CloudEvent, **kwargs) -> None:
        body = self.encoder.encode(
            message.dict(
                exclude={
                    "id",
                    "type",
                    "source",
                    "content_type",
                    "version",
                    "time",
                    "topic",
                }
            )
        )
        headers = kwargs.pop("headers", {})
        headers.setdefault("specversion", message.specversion)
        headers.setdefault("Content-Type", self.encoder.CONTENT_TYPE)
        msg = aio_pika.Message(
            headers=headers,
            body=body,
            app_id=message.source,
            content_type=message.content_type,
            timestamp=message.time,
            message_id=str(message.id),
            type=message.type,
            content_encoding="UTF-8",
            delivery_mode=aio_pika.DeliveryMode.PERSISTENT,
        )

        await self.exchange.publish(msg, routing_key=message.topic, **kwargs)

    async def _ack(self, message: aio_pika.abc.AbstractIncomingMessage) -> None:
        await message.ack()

    async def _nack(self, message: aio_pika.abc.AbstractIncomingMessage) -> None:
        await message.reject(requeue=True)

    @property
    def is_connected(self) -> bool:
        return not self.connection.is_closed

    def parse_incoming_message(
        self, message: aio_pika.abc.AbstractIncomingMessage
    ) -> Any:
        msg = self.encoder.decode(message.body)
        if not isinstance(msg, dict):
            raise TypeError(f"Expected dict, got {type(msg)}")
        msg.update(
            {
                "id": message.message_id,
                "type": message.type,
                "source": message.app_id,
                "content_type": message.content_type,
                "version": message.headers.get("specversion"),
                "time": message.timestamp,
                "topic": message.routing_key,
            }
        )
        return msg



from datetime import datetime, timedelta
from functools import partial
from typing import Any, Dict, Generic, Optional
from uuid import UUID, uuid4

from pydantic import AnyUrl, Extra, Field, validator
from pydantic.fields import PrivateAttr
from pydantic.generics import GenericModel

from .context import get_current_service
from .message import Message
from .types import D
from .utils import utc_now


class CloudEvent(GenericModel, Generic[D]):
    specversion: Optional[str] = "1.0"
    content_type: str = Field(
        "application/json", alias="datacontenttype", description="Message content type"
    )
    id: UUID = Field(default_factory=uuid4, description="Event ID")
    time: datetime = Field(default_factory=utc_now, description="Event created time")
    topic: str = Field(..., alias="subject", description="Event subject (topic)")
    type: Optional[str] = Field(None, description="Event type")
    source: Optional[str] = Field(None, description="Event source (app)")
    data: D = Field(..., description="Event payload")
    dataschema: Optional[AnyUrl] = Field(None, description="Data schema URI")
    tracecontext: Dict[str, str] = Field({}, description="Distributed tracing context")

    _raw: Optional[Any] = PrivateAttr(None)

    def __init_subclass__(cls, **kwargs):
        cls.__fields__["type"].default = cls.__name__

    def __eq__(self, other):
        if not isinstance(other, CloudEvent):
            return False
        return self.id == other.id

    @validator("type", allow_reuse=True, always=True, pre=True)
    def get_type_from_cls_name(cls, v):
        return v or cls.__name__

    @property
    def raw(self) -> Message:
        if self._raw is None:
            raise AttributeError("raw property accessible only for incoming messages")
        return self._raw

    @property
    def service(self):
        return get_current_service()

    @property
    def log_context(self) -> Dict[str, Any]:
        return {"message_id": str(self.id)}

    def _set(self, name: str, value: Any):
        object.__setattr__(self, name, value)

    def set_source(self, value: str) -> None:
        self._set("source", value)

    def set_raw(self, value: Message):
        self._set("_raw", value)

    def dict(self, **kwargs: Any) -> Dict[str, Any]:
        kwargs.setdefault("by_alias", True)
        return super().dict(**kwargs)

    @classmethod
    def new(cls, obj: D, **kwargs: Any):
        return cls(data=obj, **kwargs)

    def copy(self, **kwargs):
        kwargs.setdefault(
            "exclude",
            {
                "id",
                "time",
            },
        )
        kwargs.setdefault("deep", True)
        return super().copy(**kwargs)

    @property
    def extra_span_attributes(self) -> Dict[str, str]:
        return {}

    @property
    def age(self) -> timedelta:
        return utc_now() - self.time

    def fail(self):
        self.raw.fail()

    class Config:
        use_enum_values = True
        allow_population_by_field_name = True
        extra = Extra.allow
        arbitrary_types_allowed = True
        # events are immutable, however it's sometimes useful to set source or traceid for unpublished events
        allow_mutation = False


TopicField = partial(Field, const=True, alias="subject", description="Event topic")



from __future__ import annotations

from enum import Enum
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Callable,
    List,
    Optional,
    Protocol,
    Type,
    TypedDict,
    TypeVar,
    Union,
    runtime_checkable,
)
from uuid import UUID

if TYPE_CHECKING:
    from . import CloudEvent, GenericConsumer


ID = Union[UUID, str]


RawMessage = TypeVar("RawMessage")

T = TypeVar("T", bound="CloudEvent")
D = TypeVar("D")

Tags = Optional[List[Union[str, Enum]]]


class TagMeta(TypedDict):
    name: str
    description: str


@runtime_checkable
class Encoder(Protocol):
    """
    Encoder object protocol.
    """

    CONTENT_TYPE: str

    def encode(self, data: Any) -> bytes:
        """
        Serialize object to bytes
        :param data: input value, usually CloudEvent.dict()
        :return: raw content as bytes
        """

    def decode(self, data: bytes) -> Any:
        """
        Deserialize bytes to python object
        :param data: input bytes
        :return: de-serialized object
        """


FT = Callable[["CloudEvent"], Union[Awaitable[Optional[Any]], Optional[Any]]]
MessageHandlerT = Union[Type["GenericConsumer"], FT]

ExcHandler = Callable[["CloudEvent", Exception], Awaitable]


class ResultBackend(Protocol):
    async def get_result(self, service: str, message_id: ID) -> Any | None:
        ...



from ._version import __version__
from .asyncapi import publishes
from .broker import Broker
from .consumer import Consumer, ConsumerGroup, ForwardResponse, GenericConsumer
from .context import get_current_service
from .message import Message
from .middleware import Middleware
from .models import CloudEvent, TopicField
from .plugins import BrokerPlugin, ServicePlugin
from .runner import ServiceRunner
from .service import Service
from .types import RawMessage

__all__ = [
    "__version__",
    "Broker",
    "BrokerPlugin",
    "Consumer",
    "ConsumerGroup",
    "CloudEvent",
    "ForwardResponse",
    "GenericConsumer",
    "Message",
    "Middleware",
    "RawMessage",
    "Service",
    "ServicePlugin",
    "ServiceRunner",
    "TopicField",
    "publishes",
    "get_current_service",
]



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any

import anyio
import nats
from nats.aio.client import Client
from nats.aio.msg import Msg as NatsMsg
from nats.errors import NotJSMessageError
from nats.js import JetStreamContext
from nats.js.api import ConsumerConfig

from eventiq.broker import Broker
from eventiq.exceptions import BrokerError, PublishError

from ...message import Message
from ...utils import retry
from .settings import JetStreamSettings, NatsSettings

if TYPE_CHECKING:
    from eventiq import CloudEvent, Consumer, Service


class JsMessageProxy(Message[NatsMsg]):
    def __init__(self, message: NatsMsg):
        super().__init__(message)

    @property
    def num_delivered(self) -> int | None:
        try:
            return self._message.metadata.num_delivered
        except NotJSMessageError:
            return None


class NatsBroker(Broker[NatsMsg]):
    """
    :param url: Url to nats server(s)
    :param connection_options: additional connection options passed to nats.connect(...)
    :param auto_flush: auto flush messages on publish
    :param kwargs: options for base class
    """

    protocol = "nats"
    WILDCARD_ONE = "*"
    WILDCARD_MANY = ">"

    def __init__(
        self,
        *,
        url: str = "nats://localhost:4444",
        connection_options: dict[str, Any] | None = None,
        auto_flush: bool = True,
        **kwargs: Any,
    ) -> None:

        super().__init__(**kwargs)
        self.url = url
        self.connection_options = connection_options or {}
        self.connection_options.setdefault("pending_size", 0)
        self._auto_flush = auto_flush
        self.client = Client()

    def parse_incoming_message(self, message: NatsMsg) -> Any:
        return self.encoder.decode(message.data)

    async def _start_consumer(self, service: Service, consumer: Consumer) -> None:
        await self.client.subscribe(
            subject=self.format_topic(consumer.topic),
            queue=f"{service.name}:{consumer.name}",
            cb=self.get_handler(service, consumer),
        )

    async def _disconnect(self) -> None:
        await self.client.close()

    async def flush(self):
        await self.client.flush()

    async def _connect(self) -> None:
        await self.client.connect(self.url, **self.connection_options)

    @retry(max_retries=3)
    async def _publish(self, message: CloudEvent, **kwargs) -> None:
        data = self.encoder.encode(message.dict())
        await self.client.publish(message.topic, data, **kwargs)
        if self._auto_flush:
            await self.client.flush()

    @property
    def is_connected(self) -> bool:
        return self.client.is_connected

    Settings = NatsSettings


class JetStreamBroker(NatsBroker):
    """
    NatsBroker with JetStream enabled
    :param prefetch_count: default number of messages to prefetch
    :param fetch_timeout: timeout for subscription pull
    :param jetstream_options: additional options passed to nc.jetstream(...)
    :param kwargs: all other options for base classes NatsBroker, Broker
    """

    Settings = JetStreamSettings

    def __init__(
        self,
        *,
        prefetch_count: int = 10,
        fetch_timeout: int = 10,
        jetstream_options: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> None:

        super().__init__(**kwargs)
        self.prefetch_count = prefetch_count
        self.fetch_timeout = fetch_timeout
        self.jetstream_options = jetstream_options or {}
        self.js = JetStreamContext(self.client, **self.jetstream_options)

    @property
    def message_proxy_class(self) -> type[Message]:
        return JsMessageProxy

    @retry(max_retries=3)
    async def _publish(
        self,
        message: CloudEvent,
        **kwargs,
    ) -> None:
        data = self.encoder.encode(message)
        headers = kwargs.get("headers", {})
        timeout = kwargs.get("timeout")
        stream = kwargs.get("stream")
        headers.setdefault("Content-Type", self.encoder.CONTENT_TYPE)
        headers["Nats-Msg-Id"] = str(message.id)
        try:
            await self.js.publish(
                subject=message.topic,
                payload=data,
                timeout=timeout,
                stream=stream,
                headers=headers,
            )
            if self._auto_flush:
                await self.client.flush()
        except Exception as e:
            raise PublishError from e

    async def _start_consumer(self, service: Service, consumer: Consumer) -> None:
        durable = f"{service.name}:{consumer.name}"
        config = consumer.options.get("config", ConsumerConfig())
        config.ack_wait = consumer.timeout + 10  # consumer timeout + 10s for .ack()
        try:
            subscription = await self.js.pull_subscribe(
                subject=self.format_topic(consumer.topic),
                durable=durable,
                config=config,
            )
        except Exception as e:
            raise BrokerError(
                f"Error creating subscription for consumer {consumer.name} and topic {consumer.topic}"
            ) from e

        handler = self.get_handler(service, consumer)
        batch = consumer.options.get("prefetch_count", self.prefetch_count)
        timeout = consumer.options.get("fetch_timeout", self.fetch_timeout)
        try:
            while self._running:
                try:
                    messages = await subscription.fetch(batch=batch, timeout=timeout)
                    async with anyio.create_task_group() as tg:
                        for msg in messages:
                            tg.start_soon(handler, msg)
                except nats.errors.TimeoutError:
                    await asyncio.sleep(5)
        except Exception:
            self.logger.exception("Cancelling consumer")
        finally:
            if consumer.dynamic:
                await subscription.unsubscribe()

    async def _ack(self, message: Message) -> None:
        if not message._ackd:
            await message.ack()

    async def _nack(self, message: Message) -> None:
        if not message._ackd:
            await message.nak(delay=message.delay)



from typing import Any

import ormsgpack
from pydantic.json import pydantic_encoder

from ..exceptions import DecodeError, EncodeError


class MsgPackEncoder:
    """
    Message Pack encoder implementation using `ormsgpack` library.
    """

    CONTENT_TYPE = "application/x-msgpack"

    @staticmethod
    def encode(data: Any) -> bytes:
        try:
            return ormsgpack.packb(
                data,
                option=ormsgpack.OPT_NON_STR_KEYS | ormsgpack.OPT_SERIALIZE_NUMPY,
                default=pydantic_encoder,
            )
        except ormsgpack.MsgpackEncodeError as e:
            raise EncodeError from e

    @staticmethod
    def decode(data: bytes) -> Any:
        try:
            return ormsgpack.unpackb(data)
        except ormsgpack.MsgpackDecodeError as e:
            raise DecodeError from e



from typing import Any, Dict, Optional

from pydantic import Field

from eventiq.settings import BrokerSettings


class NatsSettings(BrokerSettings):
    url: str = Field(..., env="BROKER_URL")
    auto_flush: bool = Field(True, env="BROKER_AUTO_FLUSH")
    connection_options: Optional[Dict[str, Any]] = Field(
        None, env="BROKER_CONNECTION_OPTIONS"
    )


class JetStreamSettings(NatsSettings):
    prefetch_count: int = Field(10, env="BROKER_PREFETCH_COUNT")
    fetch_timeout: int = Field(10, env="BROKER_FETCH_TIMEOUT")
    jetstream_options: Optional[Dict[str, Any]] = Field(None, env="BROKER_OPTIONS")



from typing import Any, Dict, Optional

from pydantic import Field

from eventiq.settings import BrokerSettings


class RabbitMQSettings(BrokerSettings):
    url: str = Field(..., env="BROKER_URL")
    default_prefetch_count: int = Field(10, env="BROKER_DEFAULT_PREFETCH_COUNT")
    exchange_name: str = Field("events", env="BROKER_EXCHANGE_NAME")
    connection_options: Optional[Dict[str, Any]] = Field(None, env="CONNECTION_OPTIONS")



import signal
from typing import Sequence

import anyio

from .logger import LoggerMixin
from .service import Service


class ServiceRunner(LoggerMixin):
    """
    Service container for running multiple service instances in one process
    :param services: Sequence of services to run
    """

    def __init__(self, services: Sequence[Service]) -> None:
        self.services = services

    async def run(self):

        async with anyio.open_signal_receiver(signal.SIGINT, signal.SIGTERM) as signals:
            async with anyio.create_task_group() as tg:
                try:
                    for service in self.services:
                        tg.start_soon(service.start)
                except Exception as e:
                    self.logger.exception("Unhandled exception", exc_info=e)
                    return

                async for _ in signals:
                    self.logger.info("Exiting...")
                    await tg.cancel_scope.cancel()
                    async with anyio.move_on_after(30, shield=True):
                        for service in self.services:
                            await service.stop()
                    return



import importlib
from typing import TYPE_CHECKING, Any, Callable

from pydantic import errors
from pydantic.validators import str_validator


def import_from_string(path: str) -> Any:
    module_name, _, obj = path.partition(":")
    module = importlib.import_module(module_name)
    return getattr(module, obj)


if TYPE_CHECKING:
    ImportedType = Callable[..., Any]
else:

    class ImportedType:
        validate_always = True

        @classmethod
        def __get_validators__(cls):
            yield cls.validate

        @classmethod
        def validate(cls, value: Any) -> Any:
            if isinstance(value, Callable):
                return value

            try:
                value = str_validator(value)
            except errors.StrError as e:
                raise errors.PyObjectError(
                    error_message="value is neither a valid import path not a valid callable"
                ) from e

            try:
                return import_from_string(value)
            except ImportError as e:
                raise errors.PyObjectError(error_message=str(e)) from e



from __future__ import annotations

import asyncio
import re
from collections import defaultdict
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any

from eventiq.broker import Broker
from eventiq.middleware import Middleware

if TYPE_CHECKING:
    from eventiq import CloudEvent, Consumer, Service
    from eventiq.message import Message
    from eventiq.types import Encoder


@dataclass
class StubMessage:
    data: bytes
    queue: asyncio.Queue


class StubBroker(Broker[StubMessage]):
    """This is in-memory implementation of a broker class, mainly designed for testing."""

    protocol = "in-memory"

    WILDCARD_ONE = r"\w+"
    WILDCARD_MANY = "*"

    def __init__(
        self,
        *,
        encoder: Encoder | None = None,
        middlewares: list[Middleware] | None = None,
        **options: Any,
    ) -> None:
        super().__init__(encoder=encoder, middlewares=middlewares, **options)
        self.topics: dict[str, asyncio.Queue] = defaultdict(asyncio.Queue)
        self._stopped = False

    def parse_incoming_message(self, message: StubMessage) -> Any:
        return self.encoder.decode(message.data)

    async def _start_consumer(self, service: Service, consumer: Consumer):
        queue = self.topics[self.format_topic(consumer.topic)]
        handler = self.get_handler(service, consumer)
        while self._running:
            message = await queue.get()
            await handler(message)

    async def _connect(self) -> None:
        pass

    async def _disconnect(self) -> None:
        pass

    async def _publish(self, message: CloudEvent, **_) -> None:
        data = self.encoder.encode(message.dict())
        for topic, queue in self.topics.items():
            if re.fullmatch(topic, message.topic):
                msg = StubMessage(data=data, queue=queue)
                await queue.put(msg)

    async def _ack(self, message: Message) -> None:
        message.queue.task_done()

    async def _nack(self, message: Message) -> None:
        await message.queue.put(message)

    def is_connected(self) -> bool:  # type: ignore[override]
        return True



from __future__ import annotations

import asyncio
import inspect
from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, Callable, Generic, get_type_hints

from .logger import get_logger
from .settings import DEFAULT_TIMEOUT
from .types import FT, MessageHandlerT, T, Tags
from .utils import to_async

if TYPE_CHECKING:
    from .middlewares.retries import RetryStrategy


@dataclass(frozen=True)
class ForwardResponse:
    topic: str
    as_type: str = "CloudEvent"


class Consumer(ABC, Generic[T]):
    """Base consumer class"""

    event_type: T

    def __init__(
        self,
        *,
        topic: str,
        name: str,
        timeout: int = DEFAULT_TIMEOUT,
        dynamic: bool = False,
        forward_response: ForwardResponse | None = None,
        tags: Tags = None,
        retry_strategy: RetryStrategy | None = None,
        store_results: bool = False,
        parameters: dict[str, Any] | None = None,
        **options: Any,
    ):
        self._name = name
        self.topic = topic
        self.timeout = timeout
        self.dynamic = dynamic
        self.forward_response = forward_response
        self.tags = tags or []
        self.retry_strategy = retry_strategy
        self.store_results = store_results
        self.parameters = parameters or {}
        self.options: dict[str, Any] = options
        self.logger = get_logger(__name__, name)

    def validate_message(self, message: Any) -> T:
        return self.event_type.parse_obj(message)

    @property
    def name(self) -> str:
        return self._name

    @property
    @abstractmethod
    def description(self) -> str:
        raise NotImplementedError

    @abstractmethod
    async def process(self, message: T) -> Any | None:
        raise NotImplementedError


class FnConsumer(Consumer[T]):
    def __init__(
        self,
        *,
        topic: str,
        fn: FT,
        name: str,
        **extra: Any,
    ) -> None:
        super().__init__(name=name, topic=topic, **extra)
        event_type = get_type_hints(fn).get("message")
        assert event_type, f"Unable to resolve type hint for 'message' in {fn.__name__}"
        self.event_type = event_type
        if not asyncio.iscoroutinefunction(fn):
            fn = to_async(fn)
        self.fn = fn

    async def process(self, message: T) -> Any | None:
        return await self.fn(message)

    @property
    def description(self) -> str:
        return self.fn.__doc__ or ""


class GenericConsumer(Consumer[T], ABC):
    def __init_subclass__(cls, **kwargs):
        if not inspect.isabstract(cls):
            cls.event_type = cls.__orig_bases__[0].__args__[0]
            if not asyncio.iscoroutinefunction(cls.process):
                cls.process = to_async(cls.process)

    @property
    def description(self) -> str:
        return self.__doc__ or ""


class ConsumerGroup:
    def __init__(
        self, consumers: dict[str, Consumer] | None = None, tags: Tags = None, **options
    ):
        self.consumers = consumers or {}
        self.tags = tags or []
        self.options = options

    def add_consumer(self, consumer: Consumer) -> None:
        self.consumers[consumer.name] = consumer

    def add_consumer_group(self, other: ConsumerGroup) -> None:
        self.consumers.update(other.consumers)

    def subscribe(
        self,
        topic: str,
        *,
        name: str | None = None,
        timeout: int = DEFAULT_TIMEOUT,
        dynamic: bool = False,
        forward_response: ForwardResponse | None = None,
        tags: Tags = None,
        retry_strategy: RetryStrategy | None = None,
        store_results: bool = False,
        parameters: dict[str, Any] | None = None,
        **options,
    ) -> Callable[[MessageHandlerT], MessageHandlerT]:
        def wrapper(func_or_cls: MessageHandlerT) -> MessageHandlerT:
            cls: type[Consumer] = FnConsumer
            if isinstance(func_or_cls, type) and issubclass(
                func_or_cls, GenericConsumer
            ):
                cls = func_or_cls
                name_ = name or str(getattr(func_or_cls, "name", type(self).__name__))
            elif callable(func_or_cls):
                options["fn"] = func_or_cls
                name_ = name or func_or_cls.__name__
            else:
                raise TypeError("Expected function or GenericConsumer")
            for k, v in self.options.items():
                options.setdefault(k, v)

            consumer = cls(
                topic=topic,
                name=name_,
                forward_response=forward_response,
                timeout=timeout,
                tags=self.tags + (tags or []),
                retry_strategy=retry_strategy,
                store_results=store_results,
                parameters=parameters,
                dynamic=dynamic,
                **options,
            )

            self.add_consumer(consumer)
            return func_or_cls

        return wrapper



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .logger import LoggerMixin

if TYPE_CHECKING:
    from eventiq import Broker, CloudEvent, Consumer, Message, Service


class Middleware(LoggerMixin):
    """Base class for middlewares"""

    async def before_broker_connect(self, broker: Broker) -> None:
        """Called before broker connects"""

    async def after_broker_connect(self, broker: Broker) -> None:
        """Called after broker connects"""

    async def before_broker_disconnect(self, broker: Broker) -> None:
        """Called before broker disconnects"""

    async def after_broker_disconnect(self, broker: Broker) -> None:
        """Called after broker disconnects"""

    async def before_service_start(self, broker: Broker, service: Service):
        """Called before service starts"""

    async def after_service_start(self, broker: Broker, service: Service):
        """Called after service starts"""

    async def before_service_stop(self, broker: Broker, service: Service):
        """Called before service stops"""

    async def after_service_stop(self, broker: Broker, service: Service):
        """Called after service stops"""

    async def before_consumer_start(
        self, broker: Broker, service: Service, consumer: Consumer
    ) -> None:
        """Called before consumer is started"""

    async def after_consumer_start(
        self, broker: Broker, service: Service, consumer: Consumer
    ) -> None:
        """Called after consumer is started"""

    async def before_ack(
        self,
        broker: Broker,
        service: Service,
        consumer: Consumer,
        message: Message,
    ) -> None:
        """Called before message is acknowledged"""

    async def after_ack(
        self,
        broker: Broker,
        service: Service,
        consumer: Consumer,
        message: Message,
    ) -> None:
        """Called after message is acknowledged"""

    async def before_nack(
        self, broker: Broker, service: Service, consumer: Consumer, message: Message
    ) -> None:
        """Called before message is rejected"""

    async def after_nack(
        self, broker: Broker, service: Service, consumer: Consumer, message: Message
    ) -> None:
        """Called after message is rejected"""

    async def before_publish(
        self, broker: Broker, message: CloudEvent, **kwargs
    ) -> None:
        """Called before message is published"""

    async def after_publish(
        self, broker: Broker, message: CloudEvent, **kwargs
    ) -> None:
        """Called after message is published"""

    async def after_skip_message(
        self, broker: Broker, service: Service, consumer: Consumer, message: CloudEvent
    ) -> None:
        """Called after message is skipped by the middleware"""

    async def before_process_message(
        self, broker: Broker, service: Service, consumer: Consumer, message: CloudEvent
    ) -> None:
        """Called before message is processed"""

    async def after_process_message(
        self,
        broker: Broker,
        service: Service,
        consumer: Consumer,
        message: CloudEvent,
        result: Any | None = None,
        exc: Exception | None = None,
    ) -> None:
        """Called after message is processed (but not acknowledged/rejected yet)"""



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from eventiq.exceptions import Fail
from eventiq.middleware import Middleware

if TYPE_CHECKING:
    from eventiq import Broker, CloudEvent, Consumer, Service


class DeadLetterQueueMiddleware(Middleware):
    def __init__(self, topic: str = "dlx", type_: str = "MessageFailedEvent"):
        self.topic = topic
        self._type = type_

    async def after_process_message(
        self,
        broker: Broker,
        service: Service,
        consumer: Consumer,
        message: CloudEvent,
        result: Any | None = None,
        exc: Exception | None = None,
    ) -> None:
        if exc and isinstance(exc, Fail) or message.raw.failed:
            await service.send(self.topic, self._type, message)



from __future__ import annotations

import asyncio
import os
from pathlib import Path
from typing import TYPE_CHECKING

from eventiq.middleware import Middleware

if TYPE_CHECKING:
    from eventiq.broker import Broker


class HealthCheckMiddleware(Middleware):
    """Middleware for performing basic health checks on broker"""

    BASE_DIR = os.getenv("HEALTHCHECK_DIR", "/tmp")  # nosec

    def __init__(self, interval: int = 30, file_mode: bool = False):
        self.interval = interval
        self.file_mode = file_mode
        self._broker: Broker | None = None
        self._task: asyncio.Task | None = None

    async def after_broker_connect(self, broker: Broker):
        self._broker = broker
        if self.file_mode:
            self._task = asyncio.create_task(self._run_forever())

    async def before_broker_disconnect(self, broker: Broker):
        if self.file_mode and self._task:
            self._task.cancel()
            await self._task

    async def _run_forever(self):
        p = Path(os.path.join(self.BASE_DIR, "healthy"))
        p.touch(exist_ok=True)
        while True:
            try:
                unhealthy = not self._broker.is_connected
            except Exception as e:
                self.logger.exception("Healthcheck failed", exc_info=e)
                unhealthy = True

            if unhealthy:
                p.rename(os.path.join(self.BASE_DIR, "unhealthy"))
                break
            await asyncio.sleep(self.interval)

    def get_health_status(self) -> bool:
        if self._broker:
            return self._broker.is_connected
        return False



from __future__ import annotations

import inspect
import logging


def get_logger(module: str, name: str | type) -> logging.Logger:
    if inspect.isclass(name):
        name = name.__name__

    return logging.getLogger(f"{module}.{name}")


class LoggerMixin:
    logger: logging.Logger

    def __init_subclass__(cls, **kwargs):
        cls.logger = get_logger(__name__, cls)



from __future__ import annotations

import functools
import json
import re
from collections import defaultdict
from itertools import chain
from pathlib import Path
from typing import Iterable

from pydantic.schema import schema as all_schemas

from eventiq.asyncapi.models import (
    AsyncAPI,
    ChannelItem,
    Components,
    Info,
    Message,
    Operation,
    Parameter,
    Ref,
    Server,
    Tag,
)

from ..broker import TOPIC_PATTERN
from ..service import Service
from .registry import PUBLISH_REGISTRY

PREFIX = "#/components/schemas/"


def camel2snake(camel: str) -> str:
    """
    Converts a camelCase string to snake_case.
    """
    snake = re.sub(r"([a-zA-Z])([0-9])", lambda m: f"{m.group(1)}_{m.group(2)}", camel)
    snake = re.sub(r"([a-z0-9])([A-Z])", lambda m: f"{m.group(1)}_{m.group(2)}", snake)
    return snake.lower()


def get_all_models_schema(service: Service):
    all_models = [
        m.event_type  # type: ignore
        for m in chain(service.consumers.values(), PUBLISH_REGISTRY.values())
    ]
    return all_schemas(all_models, ref_prefix=PREFIX).get("definitions", [])


def get_topic_parameters(topic: str, **kwargs) -> dict[str, Parameter]:
    params = {}
    for k in topic.split("."):
        if TOPIC_PATTERN.fullmatch(k):
            param_name = k[1:-1]
            params[param_name] = kwargs.get(param_name, Parameter())
    return params


def get_tag_list(tags: dict[str, Tag], taggable: Iterable):
    tag_list = []
    for t in taggable:
        if t not in tags:
            tags[t] = Tag(name=t)
        tag_list.append(tags[t])
    return tag_list


def populate_spec(service: Service):
    channels: dict[str, ChannelItem] = defaultdict(ChannelItem)
    messages: dict[str, Message] = {}
    tags = {t["name"]: Tag.parse_obj(t) for t in service.tags_metadata}

    for publishes in PUBLISH_REGISTRY.values():
        event_type = publishes.event_type.__name__
        params = get_topic_parameters(publishes.topic, **publishes.kwargs)
        for k, v in params.items():
            channels[publishes.topic].parameters.setdefault(k, v)
        if event_type not in messages:
            messages[event_type] = Message(
                content_type=service.broker.encoder.CONTENT_TYPE,
                payload=Ref(ref=f"{PREFIX}{event_type}"),
                description=publishes.kwargs.get("description", ""),
            )
        channels[publishes.topic].publish = Operation(
            operation_id=f"publish_{camel2snake(event_type)}",
            message=Ref(ref=f"#/components/messages/{event_type}"),
            tags=get_tag_list(tags, publishes.kwargs.get("tags", [])),
        )
    for consumer in service.consumers.values():
        event_type = consumer.event_type.__name__
        # message_id = camel2snake(f"{consumer.name}_{event_type}")
        if event_type not in messages:
            messages[event_type] = Message(
                content_type=service.broker.encoder.CONTENT_TYPE,
                payload=Ref(ref=f"{PREFIX}{event_type}"),
                description=consumer.description,
            )
        subscribe = Operation(
            operation_id=camel2snake(consumer.name),
            message=Ref(ref=f"#/components/messages/{event_type}"),
            tags=get_tag_list(tags, consumer.tags) or None,
        )
        channels[consumer.topic].subscribe = subscribe
        params = get_topic_parameters(consumer.topic, **consumer.parameters)
        for k, v in params.items():
            channels[consumer.topic].parameters.setdefault(k, v)
    return channels, messages


@functools.lru_cache
def get_async_api_spec(service: Service) -> AsyncAPI:
    channels, messages = populate_spec(service)
    definitions = get_all_models_schema(service)
    doc_model = AsyncAPI(
        info=Info(title=service.title, version=service.version),
        servers={
            "default": Server(
                protocol=service.broker.protocol,
                description=service.broker.description,
                url=getattr(service.broker, "url", ""),
                protocol_version=service.broker.protocol_version,
            )
        },
        channels=channels,
        components=Components(schemas=definitions, messages=messages),
        default_content_type=service.broker.encoder.CONTENT_TYPE,
    )
    return doc_model


def save_async_api_to_file(spec: AsyncAPI, path: Path, fmt: str) -> None:
    dump = json.dump
    if fmt == "yaml":
        import yaml

        dump = yaml.dump
    with open(path, "w") as f:
        dump(spec.dict(by_alias=True, exclude_none=True), f)



import logging.config
import sys
from pathlib import Path
from typing import Optional

import anyio
import typer

from eventiq.logger import get_logger

from .imports import import_from_string

cli = typer.Typer()

logger = get_logger(__name__, "cli")

if "." not in sys.path:
    sys.path.insert(0, ".")


@cli.command(help="Run service")
def run(
    service_or_runner: str,
    log_level: Optional[str] = typer.Option(None),
    log_config: Optional[str] = typer.Option(None),
    use_uvloop: bool = typer.Option(False),
    debug: bool = typer.Option(False),
) -> None:
    if log_level:
        logging.basicConfig(level=log_level.upper())
    if log_config:
        logging.config.fileConfig(log_config)

    logger.info(f"Running [{service_or_runner}]...")
    obj = import_from_string(service_or_runner)
    if callable(obj):
        obj = obj()
    anyio.run(
        obj.run,
        backend="asyncio",
        backend_options={"use_uvloop": use_uvloop, "debug": debug},
    )


@cli.command(help="Watch and reload on files change")
def watch(
    service_or_runner: str = typer.Argument(...),
    log_level: str = typer.Option("INFO"),
    directory: str = typer.Option("."),
):
    from watchfiles import run_process

    logger.info(f"Watching [{service_or_runner}]...")

    run_process(
        directory,
        target=f"eventiq run {service_or_runner} --log-level={log_level}",
        target_type="command",
        callback=logger.info,
        sigint_timeout=30,
        sigkill_timeout=30,
    )


@cli.command(help="Generate AsyncAPI documentation from service")
def generate_docs(
    service: str = typer.Argument(
        ...,
        help="Global service object to import in format {package}.{module}:{service_object}",
    ),
    out: Path = typer.Option("./asyncapi.json", help="Output file path"),
    format: str = typer.Option(
        "json", help="Output format. Valid options are 'yaml' and 'json'(default)"
    ),
):
    from .asyncapi.generator import get_async_api_spec, save_async_api_to_file

    svc = import_from_string(service)
    spec = get_async_api_spec(svc)
    save_async_api_to_file(spec, out, format)
    typer.secho(f"Docs saved successfully to {out}", fg="green")



from typing import Any, Dict, Optional

from pydantic import Field

from eventiq.settings import BrokerSettings


class RedisSettings(BrokerSettings):
    url: str = Field(..., env="BROKER_URL")
    connection_options: Optional[Dict[str, Any]] = Field(
        None, env="BROKER_CONNECTION_OPTIONS"
    )



import pickle  # nosec
from typing import Any


class PickleEncoder:
    """
    Pickle encoder implementation. Allows to pass/serialize python native objects,
    but it is not recommended. See warning: <https://docs.python.org/3/library/pickle.html>
    """

    CONTENT_TYPE = "application/octet-stream"

    @staticmethod
    def encode(data: Any) -> bytes:
        return pickle.dumps(data)

    @staticmethod
    def decode(data: bytes) -> Any:
        return pickle.loads(data)  # nosec
